@@ -21,7 +21,7 @@ import urlparse
 import zipfile
 
 PROG = os.path.basename(os.path.splitext(__file__)[0])
-VERSION = '8.5.2'
+VERSION = '8.6.3'
 
 # AsciiDoc global configuration file directory.
 # NOTE: CONF_DIR is "fixed up" by Makefile -- don't rename or change syntax.
@@ -206,7 +206,6 @@ def shell(cmd, raise_error=True):
     popen.wait()
     if popen.returncode != 0 and raise_error:
         die('%s returned non-zero exit status %d' % (cmd, popen.returncode))
-#        raise subprocess.CalledProcessError(popen.returncode, cmd)
     return popen
 
 def find_resources(files, tagname, attrname, filter=None):
@@ -234,17 +233,24 @@ def find_resources(files, tagname, attrname, filter=None):
         files = [files]
     result = []
     for f in files:
-        verbose('find resources in: %s' % f)
+        verbose('finding resources in: %s' % f)
         if OPTIONS.dry_run:
             continue
         parser = FindResources()
-        parser.feed(open(f).read())
+        # HTMLParser has problems with non-ASCII strings.
+        # See http://bugs.python.org/issue3932
+        mo = re.search(r'^<\?xml.* encoding="(.*?)"', open(f).readline())
+        if mo:
+            encoding = mo.group(1)
+            parser.feed(open(f).read().decode(encoding))
+        else:
+            parser.feed(open(f).read())
         parser.close()
     result = list(set(result))   # Drop duplicate values.
     result.sort()
     return result
 
-# Not used.
+# NOT USED.
 def copy_files(files, src_dir, dst_dir):
     '''
     Copy list of relative file names from src_dir to dst_dir.
@@ -282,6 +288,44 @@ def exec_xsltproc(xsl_file, xml_file, dst_dir, opts = ''):
     finally:
         shell_cd(cwd)
 
+def get_source_options(asciidoc_file):
+    '''
+    Look for a2x command options in AsciiDoc source file.
+    Limitation: options cannot contain double-quote characters.
+    '''
+    def parse_options():
+        # Parse options to result sequence.
+        inquotes = False
+        opt = ''
+        for c in options:
+            if c == '"':
+                if inquotes:
+                    result.append(opt)
+                    opt = ''
+                    inquotes = False
+                else:
+                    inquotes = True
+            elif c == ' ':
+                if inquotes:
+                    opt += c
+                elif opt:
+                    result.append(opt)
+                    opt = ''
+            else:
+                opt += c
+        if opt:
+            result.append(opt)
+
+    result = []
+    if os.path.isfile(asciidoc_file):
+        options = ''
+        for line in open(asciidoc_file):
+            mo = re.search(r'^//\s*a2x:', line)
+            if mo:
+                options += ' ' + line[mo.end():].strip()
+        parse_options()
+    return result
+
 
 #####################################################################
 # Application class
@@ -314,12 +358,20 @@ class A2X(AttrDict):
         '''
         global ASCIIDOC
         CONF_FILE = 'a2x.conf'
+        a2xdir = os.path.dirname(os.path.realpath(__file__))
         conf_files = []
-        # From global conf directory.
-        conf_files.append(os.path.join(CONF_DIR, CONF_FILE))
         # From a2x.py directory.
-        conf_files.append(os.path.join(
-            os.path.dirname(os.path.realpath(__file__)), CONF_FILE))
+        conf_files.append(os.path.join(a2xdir, CONF_FILE))
+        # If the asciidoc executable and conf files are in the a2x directory
+        # then use the local copy of asciidoc and skip the global a2x conf.
+        asciidoc = os.path.join(a2xdir, 'asciidoc.py')
+        asciidoc_conf = os.path.join(a2xdir, 'asciidoc.conf')
+        if os.path.isfile(asciidoc) and os.path.isfile(asciidoc_conf):
+            self.asciidoc = asciidoc
+        else:
+            self.asciidoc = None
+            # From global conf directory.
+            conf_files.append(os.path.join(CONF_DIR, CONF_FILE))
         # From $HOME directory.
         home_dir = os.environ.get('HOME')
         if home_dir is not None:
@@ -329,24 +381,28 @@ class A2X(AttrDict):
             if not os.path.isfile(self.conf_file):
                 die('missing configuration file: %s' % self.conf_file)
             conf_files.append(self.conf_file)
+        # From --xsl-file option.
+        if self.xsl_file is not None:
+            if not os.path.isfile(self.xsl_file):
+                die('missing XSL file: %s' % self.xsl_file)
+            self.xsl_file = os.path.abspath(self.xsl_file)
         # Load ordered files.
         for f in conf_files:
             if os.path.isfile(f):
                 verbose('loading conf file: %s' % f)
                 execfile(f, globals())
-        # If asciidoc can't be found anywhere else look in the a2x directory.
-        if not find_executable(ASCIIDOC):
-            a2xdir = os.path.dirname(os.path.realpath(__file__))
-            asciidoc = os.path.join(a2xdir, 'asciidoc.py')
-            if os.path.isfile(asciidoc):
-                ASCIIDOC = asciidoc
+        # If asciidoc is not local to a2x then search the PATH.
+        if not self.asciidoc:
+            self.asciidoc = find_executable(ASCIIDOC)
+            if not self.asciidoc:
+                die('unable to find asciidoc: %s' % ASCIIDOC)
 
     def process_options(self):
         '''
         Validate and command options and set defaults.
         '''
         if not os.path.isfile(self.asciidoc_file):
-            die('missing input FILE: %s' % self.asciidoc_file)
+            die('missing SOURCE_FILE: %s' % self.asciidoc_file)
         self.asciidoc_file = os.path.abspath(self.asciidoc_file)
         if not self.destination_dir:
             self.destination_dir = os.path.dirname(self.asciidoc_file)
@@ -354,25 +410,38 @@ class A2X(AttrDict):
             if not os.path.isdir(self.destination_dir):
                 die('missing --destination-dir: %s' % self.destination_dir)
             self.destination_dir = os.path.abspath(self.destination_dir)
-        for d in self.resource_dirs:
-            if not os.path.isdir(d):
-                die('missing --resource-dir: %s' % d)
-        if not self.doctype:
-            if self.format == 'manpage':
-                self.doctype = 'manpage'
+        self.resource_dirs = []
+        self.resource_files = []
+        if self.resource_manifest:
+            if not os.path.isfile(self.resource_manifest):
+                die('missing --resource-manifest: %s' % self.resource_manifest)
+            for r in open(self.resource_manifest):
+                self.resources.append(r.strip())
+        for r in self.resources:
+            if os.path.isfile(r):
+                self.resource_files.append(r)
+            elif os.path.isdir(r):
+                self.resource_dirs.append(r)
             else:
-                self.doctype = 'article'
-        self.asciidoc_opts += ' --doctype %s' % self.doctype
+                die('missing resource: %s' % r)
+        # Lastly search among images and stylesheets distributed with asciidoc.
+        for p in (os.path.dirname(self.asciidoc), CONF_DIR):
+            for d in ('images','stylesheets'):
+                self.resource_dirs.append(os.path.join(p,d))
+        if not self.doctype and self.format == 'manpage':
+            self.doctype = 'manpage'
+        if self.doctype:
+            self.asciidoc_opts += ' --doctype %s' % self.doctype
         for attr in self.attributes:
             self.asciidoc_opts += ' --attribute "%s"' % attr
-        self.xsltproc_opts += ' --nonet'
+#        self.xsltproc_opts += ' --nonet'
         if self.verbose:
             self.asciidoc_opts += ' --verbose'
             self.dblatex_opts += ' -V'
         if self.icons or self.icons_dir:
             params = [
                 'callout.graphics 1',
-                'navig.graphics 0',
+                'navig.graphics 1',
                 'admon.textlabel 0',
                 'admon.graphics 1',
             ]
@@ -390,9 +459,19 @@ class A2X(AttrDict):
                 'admon.graphics 0',
             ]
         if self.stylesheet:
-            params.append('html.stylesheet "%s"' % self.stylesheet)
-        params = ['--stringparam %s' % o for o in params]
-        self.xsltproc_opts += ' ' + ' '.join(params)
+            params += ['html.stylesheet "%s"' % self.stylesheet]
+        if self.format == 'htmlhelp':
+            params += ['htmlhelp.chm "%s"' % self.basename('.chm'),
+                       'htmlhelp.hhp "%s"' % self.basename('.hhp'),
+                       'htmlhelp.hhk "%s"' % self.basename('.hhk'),
+                       'htmlhelp.hhc "%s"' % self.basename('.hhc')]
+        if self.doctype == 'book':
+            params += ['toc.section.depth 1']
+            # Books are chunked at chapter level.
+            params += ['chunk.section.depth 0']
+        for o in params:
+            if o.split()[0]+' ' not in self.xsltproc_opts:
+                self.xsltproc_opts += ' --stringparam ' + o
         if self.fop_opts:
             self.fop = True
         if os.path.splitext(self.asciidoc_file)[1].lower() == '.xml':
@@ -420,21 +499,22 @@ class A2X(AttrDict):
         Search first the directory containing the asciidoc executable then
         the global configuration file directory.
         '''
-        asciidoc = find_executable(ASCIIDOC)
-        if not asciidoc:
-            die('unable to find asciidoc: %s' % ASCIIDOC)
-        f = os.path.join(os.path.dirname(asciidoc), path)
+        f = os.path.join(os.path.dirname(self.asciidoc), path)
         if not os.path.isfile(f):
             f = os.path.join(CONF_DIR, path)
             if not os.path.isfile(f):
                 die('missing configuration file: %s' % f)
         return os.path.normpath(f)
 
-    def xsl_file(self, file_name=None):
+    def xsl_stylesheet(self, file_name=None):
         '''
         Return full path name of file in asciidoc docbook-xsl configuration
         directory.
+        If an XSL file was specified with the --xsl-file option then it is
+        returned.
         '''
+        if self.xsl_file is not None:
+            return self.xsl_file
         if not file_name:
             file_name = self.format + '.xsl'
         return self.asciidoc_conf_file(os.path.join('docbook-xsl', file_name))
@@ -446,7 +526,7 @@ class A2X(AttrDict):
         If the URIs are relative then copy them from the src_dir to the
         dst_dir.
         If not found in src_dir then recursively search all specified
-        --resource-dir's for the file name.
+        resources for the file name.
         Optional additional resources URIs can be passed in the resources list.
         Does not copy absolute URIs.
         '''
@@ -454,6 +534,7 @@ class A2X(AttrDict):
         resources += find_resources(html_files, 'link', 'href',
                         lambda attrs: attrs.get('type') == 'text/css')
         resources += find_resources(html_files, 'img', 'src')
+        resources += self.resource_files
         resources = list(set(resources))    # Drop duplicates.
         resources.sort()
         for f in resources:
@@ -491,7 +572,7 @@ class A2X(AttrDict):
                 die('missing docbook file: %s' % docbook_file)
             return
         shell('"%s" --backend docbook %s --out-file "%s" "%s"' %
-             (ASCIIDOC, self.asciidoc_opts, docbook_file, self.asciidoc_file))
+             (self.asciidoc, self.asciidoc_opts, docbook_file, self.asciidoc_file))
         if not self.no_xmllint and XMLLINT:
             shell('"%s" --nonet --noout --valid "%s"' % (XMLLINT, docbook_file))
 
@@ -500,7 +581,7 @@ class A2X(AttrDict):
         docbook_file = self.dst_path('.xml')
         xhtml_file = self.dst_path('.html')
         opts = '%s --output "%s"' % (self.xsltproc_opts, xhtml_file)
-        exec_xsltproc(self.xsl_file(), docbook_file, self.destination_dir, opts)
+        exec_xsltproc(self.xsl_stylesheet(), docbook_file, self.destination_dir, opts)
         src_dir = os.path.dirname(self.asciidoc_file)
         self.copy_resources(xhtml_file, src_dir, self.destination_dir)
 
@@ -508,7 +589,7 @@ class A2X(AttrDict):
         self.to_docbook()
         docbook_file = self.dst_path('.xml')
         opts = self.xsltproc_opts
-        exec_xsltproc(self.xsl_file(), docbook_file, self.destination_dir, opts)
+        exec_xsltproc(self.xsl_stylesheet(), docbook_file, self.destination_dir, opts)
 
     def to_pdf(self):
         if self.fop:
@@ -519,7 +600,7 @@ class A2X(AttrDict):
     def exec_fop(self):
         self.to_docbook()
         docbook_file = self.dst_path('.xml')
-        xsl = self.xsl_file('fo.xsl')
+        xsl = self.xsl_stylesheet('fo.xsl')
         fo = self.dst_path('.fo')
         pdf = self.dst_path('.pdf')
         opts = '%s --output "%s"' % (self.xsltproc_opts, fo)
@@ -552,15 +633,13 @@ class A2X(AttrDict):
         self.to_docbook()
         docbook_file = self.dst_path('.xml')
         opts = self.xsltproc_opts
-        xsl_file = self.xsl_file()
+        xsl_file = self.xsl_stylesheet()
         if self.format == 'chunked':
             dst_dir = self.dst_path('.chunked')
         elif self.format == 'htmlhelp':
             dst_dir = self.dst_path('.htmlhelp')
-            opts += ' --stringparam htmlhelp.chm "%s"' % self.basename('.chm')
-            opts += ' --stringparam htmlhelp.hhc "%s"' % self.basename('.hhc')
-            opts += ' --stringparam htmlhelp.hhp "%s"' % self.basename('.hhp')
-        opts += ' --stringparam base.dir "%s/"' % os.path.basename(dst_dir)
+        if not 'base.dir ' in opts:
+            opts += ' --stringparam base.dir "%s/"' % os.path.basename(dst_dir)
         # Create content.
         shell_rmtree(dst_dir)
         shell_makedirs(dst_dir)
@@ -571,7 +650,7 @@ class A2X(AttrDict):
 
     def to_epub(self):
         self.to_docbook()
-        xsl_file = self.xsl_file()
+        xsl_file = self.xsl_stylesheet()
         docbook_file = self.dst_path('.xml')
         epub_file = self.dst_path('.epub')
         build_dir = epub_file + '.d'
@@ -615,14 +694,17 @@ class A2X(AttrDict):
         if not self.keep_artifacts:
             shell_rmtree(build_dir)
         if self.epubcheck and EPUBCHECK:
-            shell('"%s" "%s"' % (EPUBCHECK, epub_file))
+            if not find_executable(EPUBCHECK):
+                warning('epubcheck skipped: unable to find executable: %s' % EPUBCHECK)
+            else:
+                shell('"%s" "%s"' % (EPUBCHECK, epub_file))
 
     def to_text(self):
         text_file = self.dst_path('.text')
         html_file = self.dst_path('.text.html')
         if self.lynx:
             shell('"%s" %s --conf-file "%s" -b html4 -o "%s" "%s"' %
-                 (ASCIIDOC, self.asciidoc_opts, self.asciidoc_conf_file('text.conf'),
+                 (self.asciidoc, self.asciidoc_opts, self.asciidoc_conf_file('text.conf'),
                   html_file, self.asciidoc_file))
             shell('"%s" -dump "%s" > "%s"' %
                  (LYNX, html_file, text_file))
@@ -630,9 +712,8 @@ class A2X(AttrDict):
             # Use w3m(1).
             self.to_docbook()
             docbook_file = self.dst_path('.xml')
-            xhtml_file = self.dst_path('.html')
             opts = '%s --output "%s"' % (self.xsltproc_opts, html_file)
-            exec_xsltproc(self.xsl_file(), docbook_file,
+            exec_xsltproc(self.xsl_stylesheet(), docbook_file,
                     self.destination_dir, opts)
             shell('"%s" -cols 70 -dump -T text/html -no-graph "%s" > "%s"' %
                  (W3M, html_file, text_file))
@@ -647,14 +728,14 @@ class A2X(AttrDict):
 if __name__ == '__main__':
     description = '''A toolchain manager for AsciiDoc (converts Asciidoc text files to other file formats)'''
     from optparse import OptionParser
-    parser = OptionParser(usage='usage: %prog [OPTIONS] FILE',
+    parser = OptionParser(usage='usage: %prog [OPTIONS] SOURCE_FILE',
         version='%s %s' % (PROG,VERSION),
         description=description)
     parser.add_option('-a', '--attribute',
         action='append', dest='attributes', default=[], metavar='ATTRIBUTE',
         help='set asciidoc attribute value')
     parser.add_option('--asciidoc-opts',
-        action='store', dest='asciidoc_opts', default='',
+        action='append', dest='asciidoc_opts', default=[],
         metavar='ASCIIDOC_OPTS', help='asciidoc options')
     #DEPRECATED
     parser.add_option('--copy',
@@ -665,7 +746,7 @@ if __name__ == '__main__':
         help='configuration file')
     parser.add_option('-D', '--destination-dir',
         action='store', dest='destination_dir', default=None, metavar='PATH',
-        help=' output directory (defaults to FILE directory)')
+        help='output directory (defaults to SOURCE_FILE directory)')
     parser.add_option('-d','--doctype',
         action='store', dest='doctype', metavar='DOCTYPE',
         choices=('article','manpage','book'),
@@ -697,10 +778,18 @@ if __name__ == '__main__':
     parser.add_option('-n','--dry-run',
         action='store_true', dest='dry_run', default=False,
         help='just print the commands that would have been executed')
-    parser.add_option('-r','--resource-dir',
-        action='append', dest='resource_dirs', default=[],
+    parser.add_option('-r','--resource',
+        action='append', dest='resources', default=[],
         metavar='PATH',
-        help='directory containing images and stylesheets')
+        help='resource file or directory containing resource files')
+    parser.add_option('-m', '--resource-manifest',
+        action='store', dest='resource_manifest', default=None, metavar='FILE',
+        help='read resources from FILE')
+    #DEPRECATED
+    parser.add_option('--resource-dir',
+        action='append', dest='resources', default=[],
+        metavar='PATH',
+        help='DEPRECATED: use --resource')
     #DEPRECATED
     parser.add_option('-s','--skip-asciidoc',
         action='store_true', dest='skip_asciidoc', default=False,
@@ -714,28 +803,38 @@ if __name__ == '__main__':
         action='store_true', dest='safe', default=False,
         help='DEPRECATED: does nothing')
     parser.add_option('--dblatex-opts',
-        action='store', dest='dblatex_opts', default='',
+        action='append', dest='dblatex_opts', default=[],
         metavar='DBLATEX_OPTS', help='dblatex options')
     parser.add_option('--fop',
         action='store_true', dest='fop', default=False,
         help='use FOP to generate PDF files')
     parser.add_option('--fop-opts',
-        action='store', dest='fop_opts', default='',
+        action='append', dest='fop_opts', default=[],
         metavar='FOP_OPTS', help='options for FOP pdf generation')
     parser.add_option('--xsltproc-opts',
-        action='store', dest='xsltproc_opts', default='',
+        action='append', dest='xsltproc_opts', default=[],
         metavar='XSLTPROC_OPTS', help='options for FOP pdf generation')
+    parser.add_option('--xsl-file',
+        action='store', dest='xsl_file', metavar='XSL_FILE',
+        help='custom XSL stylesheet')
     parser.add_option('-v', '--verbose',
         action='count', dest='verbose', default=0,
         help='increase verbosity')
     if len(sys.argv) == 1:
         parser.parse_args(['--help'])
-    opts, args = parser.parse_args()
+    source_options = get_source_options(sys.argv[-1])
+    argv = source_options + sys.argv[1:] 
+    opts, args = parser.parse_args(argv)
     if len(args) != 1:
         parser.error('incorrect number of arguments')
+    opts.asciidoc_opts = ' '.join(opts.asciidoc_opts)
+    opts.dblatex_opts = ' '.join(opts.dblatex_opts)
+    opts.fop_opts = ' '.join(opts.fop_opts)
+    opts.xsltproc_opts = ' '.join(opts.xsltproc_opts)
     opts = eval(str(opts))  # Convert optparse.Values to dict.
     a2x = A2X(opts)
     OPTIONS = a2x           # verbose and dry_run used by utility functions.
+    verbose('args: %r' % argv)
     a2x.asciidoc_file = args[0]
     try:
         a2x.load_conf()
